#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rgrid.h"

extern void *scalloc(size_t nmemb, size_t size);

RGrid2D::RGrid2D(int w, int l, double cls, double **s)
{
   int i, j;
   
   Width = w;
   Length = l;
   CLengthSqu = cls;
   Sources = s;
   
   Nodes = (double **)scalloc(Width, sizeof(double *));
   HConnected = (int **)scalloc(Width - 1, sizeof(int *));
   VConnected = (int **)scalloc(Width, sizeof(int *));
   HCurr = (double **)scalloc(Width - 1, sizeof(double *));
   VCurr = (double **)scalloc(Width, sizeof(double *));
   HDiff = (double **)scalloc(Width - 1, sizeof(double *));
   VDiff = (double **)scalloc(Width, sizeof(double *));

   for (i = 0; i < Width; i++)
     {
	Nodes[i] = (double *)scalloc(Length, sizeof(double));
	VConnected[i] = (int *)scalloc(Length - 1, sizeof(int));
	for (j = 0; j < Length - 1; j++)
	  VConnected[i][j] = 1;
	VCurr[i] = (double *)scalloc(Length - 1, sizeof(double));
	VDiff[i] = (double *)scalloc(Length - 1, sizeof(double));
	if (i != Width)
	  {
	     HConnected[i] = (int *)scalloc(Length, sizeof(int));
	     for (j = 0; j < Length; j++)
	       HConnected[i][j] = 1;
	     HCurr[i] = (double *)scalloc(Length, sizeof(double));
	     HDiff[i] = (double *)scalloc(Length, sizeof(double));
	  }
     }
}

RGrid2D::~RGrid2D()
{
   int i;

   for (i = 0; i < Width; i++)
     {
	free(Nodes[i]);
	free(VConnected[i]);
	free(VCurr[i]);
	free(VDiff[i]);
	if (i != Width)
	  {
	     free(HConnected[i]);
	     free(HCurr[i]);
	     free(HDiff[i]);
	  }
     }
   free(Nodes);
   free(HConnected);
   free(VConnected);
   free(HCurr);
   free(VCurr);
   free(HDiff);
   free(VDiff);
}

void RGrid2D::InitialiseNodes()
{
   int i, j;
   
   for (i = 0; i < Width; i++)
     {
	for (j = 0; j < Length; j++)
	  Nodes[i][j] = Sources[i][j];
     }
}

void RGrid2D::IterateH(int i, int j, float Increment)
{
   double newcurr;
   
   HDiff[i][j] = Nodes[i][j] - Nodes[i + 1][j];
   if (HConnected[i][j])
     newcurr = HDiff[i][j];
   else
     newcurr = 0.0;
   HCurr[i][j] = HCurr[i][j] * (1.0 - Increment) + newcurr * Increment;
}

void RGrid2D::IterateV(int i, int j, float Increment)
{
   double newcurr;
   
   VDiff[i][j] = Nodes[i][j] - Nodes[i][j + 1];
   if (VConnected[i][j])
     newcurr = VDiff[i][j];
   else
     newcurr = 0.0;
   VCurr[i][j] = VCurr[i][j] * (1.0 - Increment) + newcurr * Increment;
}

void RGrid2D::CalcNode(int i, int j, float Increment)
{
   double curr = 0.0;
   double newnode;
   double change = Nodes[i][j];

   if (i != 0)
     curr += HCurr[i - 1][j];
   if (i != Width - 1)
     curr -= HCurr[i][j];
   if (j != 0)
     curr += VCurr[i][j - 1];
   if (j != Length - 1)
     curr -= VCurr[i][j];
   newnode = Sources[i][j] + curr / CLengthSqu;
   Nodes[i][j] = Nodes[i][j] * (1.0 - Increment) + newnode * Increment;
   change -= Nodes[i][j];
   if (fabs(change) >= LargestChange)
     LargestChange = fabs(change);
}

double RGrid2D::Iterate(float Increment)
{
   int i, j;

   LargestChange = 0.0;
   for (i = 0; i < Width - 1; i++)
     {
	for (j = 0; j < Length; j++)
	  IterateH(i, j, Increment);
     }
   for (i = 0; i < Width; i++)
     {
	for (j = 0; j < Length - 1; j++)
	  IterateV(i, j, Increment);
     }
   for (i = 0; i < Width; i++)
     {
	for (j = 0; j < Length; j++)
	  CalcNode(i, j, Increment);
     }
   return LargestChange;
}

void RGrid2D::PrintNodeTable()
{
   int i, j;
   
   for (j = 0; j < Length; j++)
     {
	for (i = 0; i < Width; i++)
	  printf("%.2f ", Nodes[i][j]);
	printf("\n");
     }
}

void RGrid2D::PrintHCurrTable()
{
   int i, j;
   
   for (j = 0; j < Length; j++)
     {
	for (i = 0; i < Width - 1; i++)
	  printf("%.2f ", HCurr[i][j]);
	printf("\n");
     }
}

void RGrid2D::PrintVCurrTable()
{
   int i, j;
   
   for (j = 0; j < Length - 1; j++)
     {
	for (i = 0; i < Width; i++)
	  printf("%.2f ", VCurr[i][j]);
	printf("\n");
     }
}

void RGrid2D::PrintSegmentation()
{
   int i, j;
   
   for (j = 0; j < Length - 1; j++)
     {
	for (i = 0; i < Width - 1; i++)
	  printf(HConnected[i][j] ? " ---":"    ");
	printf("\n");
	for (i = 0; i < Width - 1; i++)
	  {
	     printf(VConnected[i][j] ? "|":" ");
	     printf("   ");
	  }
	printf(VConnected[Width - 1][j] ? "|\n":" \n");
     }
   for (i = 0; i < Width - 1; i++)
     printf(HConnected[i][Length - 1] ? " ---":"    ");
   printf("\n");
}

void RGrid2D::PrintHeader(FILE *f, const char *hfile)
{
   FILE *header;
   char c;

   fprintf(f, "%%!PS-Adobe-2.0\n");
   fprintf(f, "%%%%BoundingBox: 0 0 %d %d\n\n", Width, Length);
   header = fopen(hfile, "r");
   while ((c = getc(header)) != EOF)
     putc(c, f);
   fclose(header);
}

void RGrid2D::PrintNodes(FILE *f)
{
   int i, j;

   for (i = 0; i < Width; i++)
     {
	for (j = 0; j < Length; j++)
	  fprintf(f, "%3d %3d %.3f pixel\n", i, j, (Nodes[i][j] - 1.5) / 2.0);
     }
}

void RGrid2D::PrintPostscript(const char *psfile)
{
   FILE *f = fopen(psfile, "w");

   PrintHeader(f, "header.ps");
   PrintNodes(f);
   fclose(f);
}

void RGrid2D::PrintEdgesPostscript(const char *psfile)
{
   FILE *f = fopen(psfile, "w");
   int i, j;

   PrintHeader(f, "header.ps");
   PrintNodes(f);
   for (i = 0; i < Width; i++)
     {
	for (j = 0; j < Length - 1; j++)
	  {
	     if (VConnected[i][j] == 0)
	       fprintf(f, "%d %d %d %d dotline\n", i, j + 1, i + 1, j + 1);
	  }
     }
   for (i = 0; i < Width - 1; i++)
     {
	for (j = 0; j < Length; j++)
	  {
	     if (HConnected[i][j] == 0)
	       fprintf(f, "%d %d %d %d dotline\n", i + 1, j, i + 1, j + 1);
	  }
     }
   fclose(f);
}

